@@ -8,7 +8,6 @@ import  io
 from    atomicwrites import atomic_write
 from    more_itertools import peekable
 
-import  zipfile
 from    contextlib import nullcontext
 
 from    .           import cells as c, notebook
@@ -23,6 +22,8 @@ except ImportError:
     root = True
     using_mpi = False
 
+skip_repl = False
+
 # workaround for a bug in OpenMPI (or anything else that screws up the terminal size);
 # see https://github.com/willmcgugan/rich/issues/127
 import  shutil
@@ -55,9 +56,6 @@ def show_console(cell, rule = False, verbose = False, no_show = False):
     if not no_show:
         cell.show_console(console)
 
-    if not cell.empty() and not rule:
-        console.print()
-
 def show_html(cell, f):
     cell.show_html(f)
 
@@ -91,11 +89,7 @@ def show(fn: "input notebook",
          debug: "show debugging information" = False):
     """Show the contents of the notebook, without evaluating."""
     with open(fn) as f:
-        if external and not os.path.exists(external):
-            warn(f"External zip archive [error]{external}[/error] not found. Ignoring.")
-            external = ''
-        with zipfile.ZipFile(external, 'r') if external else nullcontext() as external_zip:
-            cells = c.parse(f, external_zip, show_only = True, info=info)
+        cells = c.parse(f, external, show_only = True, info=info)
     _show(cells, html, katex, debug)
 
 def _show(cells, html, katex, debug):
@@ -126,6 +120,7 @@ def _show(cells, html, katex, debug):
         f_html.write('</body>\n')
         f_html.write('</html>\n')
 
+@argh.arg('infn', nargs='?')
 @argh.arg('outfn', nargs='?')
 @argh.arg('-n', '--dry-run')
 @argh.arg('-i', '--interactive')
@@ -139,21 +134,18 @@ def run(infn: "input notebook",
         only_root_output: "suppress output everywhere but rank 0 (for MPI)" = False,
         interactive: "run REPL after the notebook is processed" = False):
     """Run the notebook."""
-    if os.path.exists(infn):
+    if infn and os.path.exists(infn):
         with open(infn) as f:
-            if external and os.path.exists(external):
-                with zipfile.ZipFile(external, 'r') if external else nullcontext() as external_zip:
-                    cells = c.parse(f, external_zip, info=info)
-            else:
-                if external:
-                    warn(f"External zip archive [error]{external}[/error] not found. Will use it for output only.")
-                cells = c.parse(f, '', info=info)
+            cells = c.parse(f, external, info=info)
     else:
         cells = []
         if outfn and not dry_run:
             warn(f"Input file [error]{infn}[/error] doesn't exist, but given an output file [cyan]{outfn}[/cyan]; forcing [affirm]--dry-run[/affirm]")
             dry_run = True
 
+    if not infn:
+        interactive = True
+
     if not outfn:
         outfn = infn
 
@@ -163,7 +155,7 @@ def run(infn: "input notebook",
         if root or (not only_root_output and type(cell) is c.OutputCell):
             show_console(cell, rule = debug, verbose = debug)
 
-    nb = notebook.Notebook(name = infn, auto_capture = auto_capture)
+    nb = notebook.Notebook(name = infn, auto_capture = auto_capture, dry_run = dry_run)
     nb.add(cells)
 
     if not clean:
@@ -174,18 +166,32 @@ def run(infn: "input notebook",
             info('Resuming', style="magenta")
 
     try:
-        nb.process(output, force=interactive, info=info, debug=debug)
+        nb.process_all(output,
+                       repl=lambda: run_repl(nb, output, debug=debug,
+                                             prefix = [c.Blanks.create(1)], suffix = [c.Blanks.create(1), c.BreakCell.create()]),
+                       force=interactive, info=info, debug=debug)
 
         if interactive:
-            run_repl(nb, output, outfn, external, dry_run, debug=debug)
+            result = run_repl(nb, output, debug=debug)
+            if result and not nb.dry_run and not outfn:
+                from prompt_toolkit import prompt
+                from prompt_toolkit.completion import PathCompleter
+                outfn = prompt("Notebook filename (empty to not save): ", completer = PathCompleter())
+                if outfn and not external:
+                    external = prompt("External zip archive filename (empty to inline): ", completer = PathCompleter())
     except:
         nb.move_all_incoming()
 
-    if not dry_run and root:
+    if not nb.dry_run and root and outfn:
         nb.save(outfn, external)
 
 
-def run_repl(nb, output, outfn = '', external = '', dry_run = True, debug = False):
+def run_repl(nb, output, debug = False,
+             prefix = [c.Blanks.create(1), c.BreakCell.create(), c.Blanks.create(1)],
+             suffix = []):
+    if skip_repl:
+        return
+
     if using_mpi:
         comm = MPI.COMM_WORLD
 
@@ -194,13 +200,13 @@ def run_repl(nb, output, outfn = '', external = '', dry_run = True, debug = Fals
         if using_mpi and root:
             line = comm.bcast(line, root = 0)
         cells = []
-        if len(nb) > 0:
-            cells.append(c.Blanks.create(1))
-            cells.append(c.BreakCell())
-            cells.append(c.Blanks.create(1))
+        if nb.current > 0:
+            cells += prefix
         cells += c.parse(io.StringIO(line), None, info=info)
-        nb.add(cells)
-        nb.process(output,info=info)
+        cells += suffix
+
+        nb.insert(cells)
+        nb.process_to(nb.current + len(cells),output,info=info)
 
     if not root:
         while True:
@@ -212,32 +218,61 @@ def run_repl(nb, output, outfn = '', external = '', dry_run = True, debug = Fals
     if not os.path.exists(saturn_dir):
         os.makedirs(saturn_dir)
 
-    repl = PythonReplWithExecute(
-        execute = execute_line,
-        # get_globals=lambda: nb.g,     # this breaks checkpointing inside repl;
-                                        # not sure why, but commenting it out for now
-        get_locals=lambda: nb.l,
-        vi_mode=False,
-        history_filename=saturn_dir + '/history',
-        startup_paths=None,
-        debug=debug,
-    )
+    try:
+        g = {}  # fake empty globals;
+                # not specifying any get_globals throws an exception inside repl.run();
+                # specifying nb.globals breaks checkpointing inside repl; not sure why
+        repl = PythonReplWithExecute(
+            execute = execute_line,
+            get_globals=lambda: g,
+            get_locals=lambda: nb.l,
+            vi_mode=False,
+            history_filename=saturn_dir + '/history',
+            startup_paths=None,
+            debug=debug,
+        )
+    except Exception as e:
+        print("Caught exception setting up REPL:", e)
+        raise
 
     # Add the code cells to history
     for cell in nb.cells:
         for line in cell.repl_history():
             repl.history.append_string(line)
 
+    result = True
+
+    @repl.add_key_binding('c-q')
+    def _(event):
+        nonlocal result
+        event.app.exit(exception=EOFError)
+        result = False
+
+    # TODO: think about changing this to a toggle
+    @repl.add_key_binding('f10')
+    def _(event):
+        nonlocal result
+        nb.dry_run = True
+        event.app.exit(exception=EOFError)
+        result = False
+
     @repl.add_key_binding('c-w')
     def _(event):
-        if not dry_run and root:
-            nb.save(outfn, external)
+        global skip_repl
+        skip_repl = True
+        event.app.exit(exception=EOFError)
 
-    repl.run()
+    try:
+        repl.run()
+    except Exception as e:
+        print("Caught exception in REPL:", e)
+        raise
 
     if using_mpi:
         comm.bcast('', root = 0)
 
+    return result
+
 @argh.arg('outfn', nargs='?')
 def clean(infn: "input notebook",
           outfn: "output notebook (if empty, input modified in place)",
@@ -275,8 +310,7 @@ def image(infn: "input notebook", i: "image index", out: "output PNG filename",
         return
 
     with open(infn) as f:
-        with zipfile.ZipFile(external, 'r') if external else nullcontext() as external_zip:
-            cells = c.parse(f, external_zip, show_only = True, info=info)
+        cells = c.parse(f, external, show_only = True, info=info)
 
     count = 0
     for cell in cells:
@@ -330,7 +364,7 @@ def convert(infn: "Jupyter notebook",
         elif jcell['cell_type'] == 'code':
             if type(cells[-1]) is c.CodeCell:
                 cells.append(c.Blanks.create(1))
-                cells.append(c.BreakCell())
+                cells.append(c.BreakCell.create())
                 cells.append(c.Blanks.create(1))
             cell = c.CodeCell()
             cell.lines_ = [line + '\n' for line in jcell['source'].split('\n')]
@@ -374,8 +408,7 @@ def rehash(infn: "input notebook",
         outfn = infn
 
     with open(infn) as f:
-        with zipfile.ZipFile(external, 'r') if external else nullcontext() as external_zip:
-            cells = c.parse(f, external_zip, info=info)
+        cells = c.parse(f, external, info=info)
 
     nb = notebook.Notebook(name = infn)
     nb.add(cells)
@@ -409,8 +442,7 @@ def embed(infn: "input notebook",
         outfn = infn
 
     with open(infn) as f:
-        with zipfile.ZipFile(external, 'r') if external else nullcontext() as external_zip:
-            cells = c.parse(f, external_zip, info=info)
+        cells = c.parse(f, external, info=info)
 
     nb = notebook.Notebook(name = infn)
     nb.add(cells)
