@@ -22,6 +22,8 @@ from . import utils
 from . import image
 from . import evaluate
 
+import  zipfile, os
+
 # Prefix for the filename inside an external zipfile
 _zip_fn_prefix = 'name='
 
@@ -154,8 +156,15 @@ class OutputCell(Cell):
                     # take everything from name= to the end (-1 to not include \n)
                     fn = line[line.index(_zip_fn_prefix) + len(_zip_fn_prefix):-1]
                     if external:
-                        png_content = external.read(fn)
-                        self.composite_.append_png(png_content)
+                        if zipfile.Path(external, at=fn).exists():
+                            png_content = external.read(fn)
+                            self.composite_.append_png(png_content)
+                        else:
+                            # here and below, we use
+                            # self.composite_.append_rich, instead of info, so
+                            # that the message appears in place of the cell,
+                            # rather than before any other cell is processed
+                            self.composite_.append_rich(Rule(f"[warn]image hash [cyan]{fn}[/cyan] not found in the external archive [cyan]{external.filename}[/cyan][/warn]"))
                     else:
                         self.composite_.append_rich(Rule(f"[warn]image hash found, but no external file given ([cyan]{fn}[/cyan])[/warn]"))
                 else:
@@ -215,8 +224,37 @@ class BreakCell(Cell):
     def display(cls):
         return False
 
-    def save(self, external):
-        return [self._prefix + '\n']
+    @staticmethod
+    def create():
+        cell = BreakCell()
+        cell.lines_ = ['\n']
+        return cell
+
+
+class REPLCell(Cell):
+    _prefix = '#-REPL-#'
+
+    @classmethod
+    def display(cls):
+        return False
+
+class SaturnCell(Cell):
+    _prefix = '#saturn>'
+
+    _external_prefix = 'external='
+
+    def parse(self, external, info):            # called after all the lines have been read
+        super().parse(external, info)
+        line = self.lines_[0]
+        self.external_fn = ''
+        if self._external_prefix in line:
+            # take everything from external= to the end (-1 to not include \n)
+            # eventually, we'll want a more elaborate support, once we have other kinds of metadata
+            self.external_fn = line[line.index(self._external_prefix) + len(self._external_prefix):-1]
+
+    @classmethod
+    def display(cls):
+        return False
 
 class CheckpointCell(Cell):
     _prefix = '#chk>'
@@ -250,7 +288,11 @@ class CheckpointCell(Cell):
         if _zip_fn_prefix in line:
             fn = line[line.index(_zip_fn_prefix) + len(_zip_fn_prefix):-1]
             if external:
-                self._content = io.BytesIO(external.read(fn))
+                if zipfile.Path(external, at=fn).exists():
+                    self._content = io.BytesIO(external.read(fn))
+                else:
+                    self._warning = Rule(f"[warn]{self._warning_name} hash [cyan]{fn}[/cyan] not found in the external archive [cyan]{external.filename}[/cyan][/warn]")
+                    self._expected = None
             else:
                 self._warning = Rule(f"[warn]{self._warning_name} hash found, but no external file given ([cyan]{fn}[/cyan])[/warn]")
                 self._expected = None
@@ -350,7 +392,7 @@ class Blanks(Cell):
         cell.lines_ = ['\n']*n
         return cell
 
-cell_types = [MarkdownCell, OutputCell, BreakCell, CheckpointCell, VariableCell]
+cell_types = [MarkdownCell, OutputCell, BreakCell, CheckpointCell, VariableCell, REPLCell, SaturnCell]
 
 def identify(line):
     for Type in chain(cell_types, [CodeCell]):        # CodeCell matches everything, so comes last
@@ -363,11 +405,35 @@ def chunk(content, width, markers = False):
         chunking = chain(['{{{'], chunking, ['}}}'])
     return chunking
 
-def parse(f, external, show_only = False, *, info = lambda *args: None):
+def open_external(external_fn, show_only, info):
+    if external_fn:
+        if not os.path.exists(external_fn):
+            if show_only:
+                info(f"External zip archive [error]{external_fn}[/error] not found.", style='warn')
+            return None
+
+        return zipfile.ZipFile(external_fn, 'r')
+    else:
+        return None
+
+def parse(f, external_fn, *, show_only = False, info = lambda *args, **kwargs: None):
+    external = open_external(external_fn, show_only, info)
+
+    def should_parse(cell):
+        # skip check-point cells in show mode
+        return not show_only or cell.display() or type(cell) is SaturnCell
+
     cells = []
     def cells_append(cell):
-        if len(cells) > 0:
+        nonlocal external, external_fn
+        if len(cells) > 0 and should_parse(cells[-1]):
             cells[-1].parse(external, info)
+
+            if type(cells[-1]) is SaturnCell:
+                if not external_fn and cells[-1].external_fn:
+                    external_fn = cells[-1].external_fn
+                    external = open_external(external_fn, show_only, info)
+
         cells.append(cell)
 
     p = peekable(f)
@@ -389,13 +455,12 @@ def parse(f, external, show_only = False, *, info = lambda *args: None):
 
         Type = identify(line)
 
-        if show_only and not Type.display(): continue     # skip check-point cells in show mode
-
         if len(cells) == 0 or type(cells[-1]) is not Type:
             cells_append(Type())
         cells[-1].append(line)
 
     if len(cells) > 0:
-        cells[-1].parse(external, info)
+        if should_parse(cells[-1]):
+            cells[-1].parse(external, info)
 
     return cells
