@@ -25,10 +25,11 @@ class Hasher:
         return self.m.digest()
 
 class Notebook:
-    def __init__(self, name = '', auto_capture = False, debug = False):
+    def __init__(self, *, name = '', auto_capture = False, debug = False, dry_run = False):
         self.name = name
         self.auto_capture = auto_capture
         self.debug = debug
+        self.dry_run = dry_run
 
         self.incoming = []
         self.current  = 0
@@ -45,6 +46,9 @@ class Notebook:
     def add(self, cells):
         self.incoming += cells
 
+    def insert(self, cells):
+        self.incoming = self.incoming[:self.current] + cells + self.incoming[self.current:]
+
     def skip(self, location, output = lambda x: None):
         while self.current < location:
             cell = self.incoming[self.current]
@@ -141,8 +145,14 @@ class Notebook:
         if type(self.next_cell()) is c.OutputCell:
             self.current += 1
 
-    def process(self, output, *, force = False, debug = False, info = lambda *args, **kwargs: None):
+    def process_all(self, output, **kwargs):
+        # Since extra cells may be added via REPL during the execution,
+        # we need the outer loop to make sure everything is processed
         while self.current < len(self.incoming):
+            self.process_to(len(self.incoming), output, **kwargs)
+
+    def process_to(self, to, output, *, repl = lambda: None, force = False, debug = False, info = lambda *args, **kwargs: None):
+        while self.current < to:
             cell = self.incoming[self.current]
             self.current += 1
 
@@ -183,6 +193,9 @@ class Notebook:
                     tb = Traceback(self, self.debug)
                     info(tb, block = True)
                     info(f"[affirm]continuing[/affirm]")
+            elif type(cell) is c.REPLCell:
+                if repl() == False:
+                    raise SystemExit
 
     def append(self, cell, output = lambda x: None):
         self.cells.append(cell)
@@ -202,6 +215,15 @@ class Notebook:
             self.append(cell)
 
     def save(self, fn, external):
+        if self.dry_run:
+            return
+
+        if not external:
+            for cell in self.cells:
+                if type(cell) is c.SaturnCell and cell.external_fn:
+                    external = cell.external_fn
+                    break
+
         with atomic_write(fn, mode='w', overwrite=True) as f:
             with zipfile.ZipFile(external, 'w') if external else nullcontext() as external_zip:
                 for i,cell in enumerate(self.cells):
